use serde::{
    de::{self, Deserializer},
    Deserialize, Serialize,
};
use thiserror::Error;

pub mod client;
pub mod types;

#[derive(Debug, Error)]
pub enum Error {}

#[derive(Serialize, Debug, Clone)]
pub enum Message {
    Request(client::Request),
    Close(types::SubscriptionId),
    Notice(String),
}

impl Message {
    pub fn get_request(&self) -> Option<&client::Request> {
        match self {
            Self::Request(x) => Some(x),
            _ => None,
        }
    }

    pub fn get_notice(&self) -> Option<&str> {
        match self {
            Self::Notice(x) => Some(x),
            _ => None,
        }
    }

    pub fn get_close_subscription_id(&self) -> Option<&types::SubscriptionId> {
        match self {
            Self::Close(x) => Some(x),
            _ => None,
        }
    }
}

impl<'de> de::Deserialize<'de> for Message {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        let array: Vec<serde_json::Value> = Deserialize::deserialize(deserializer)?;
        if array.is_empty() {
            return Err(de::Error::custom(
                "Invalid array length, expecting at least one",
            ));
        }

        let tag = array
            .get(0)
            .ok_or_else(|| de::Error::custom("Invalid array length, expecting at least one"))?
            .as_str()
            .ok_or_else(|| de::Error::custom("Invalid type for element 0 of the array"))?;

        match tag {
            "NOTICE" => Ok(Self::Notice(
                serde_json::from_value(
                    array
                        .get(1)
                        .ok_or_else(|| de::Error::custom("Missing element 1 of the array"))?
                        .clone(),
                )
                .map_err(|e: serde_json::Error| de::Error::custom(e.to_string()))?,
            )),
            "CLOSE" => Ok(Self::Close(
                serde_json::from_value(
                    array
                        .get(1)
                        .ok_or_else(|| de::Error::custom("Missing element 1 of the array"))?
                        .clone(),
                )
                .map_err(|e: serde_json::Error| de::Error::custom(e.to_string()))?,
            )),
            "REQ" => {
                let subscription_id = array
                    .get(1)
                    .ok_or_else(|| de::Error::custom("Missing element 1 in the array"))?
                    .as_str()
                    .ok_or_else(|| {
                        de::Error::custom("Invalid type for element 1, expecting a string")
                    })?
                    .try_into()
                    .map_err(|e: types::subscription_id::Error| {
                        de::Error::custom(format!("Invalid subscription id: {}", e))
                    })?;

                Ok(Self::Request(client::Request {
                    subscription_id,
                    filters: if array.len() > 2 {
                        serde_json::from_value::<Vec<types::Filter>>(serde_json::Value::Array(
                            array[2..].to_owned(),
                        ))
                        .map_err(|e: serde_json::Error| de::Error::custom(e.to_string()))?
                    } else {
                        vec![]
                    },
                }))
            }
            tag => Err(de::Error::custom(format!("{} is not a support tag", tag))),
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn unsupported() {
        let json = "[\"CLOSEX\", \"foo\"]";
        let message: Result<Message, _> = serde_json::from_str(json);
        assert!(message.is_err());
    }

    #[test]
    fn close() {
        let json = "[\"CLOSE\", \"foo\"]";
        let message: Message = serde_json::from_str(json).expect("valid message");
        let subscription_id = message
            .get_close_subscription_id()
            .expect("valid subscription_id");

        assert_eq!("foo".to_owned(), subscription_id.to_string());
    }
}
