use super::Addr;
use serde::{
    de::{self, Deserialize, Deserializer},
    ser::{self, SerializeSeq, Serializer},
};
use thiserror::Error;

#[derive(Error, Debug)]
pub enum Error {
    #[error("Subscription ID is limited to 64 characters")]
    TooLong,
}

#[derive(Debug, PartialEq, Eq, Clone)]
pub enum TagType {
    Event,
    PubKey,
    Custom(String),
}

impl From<&str> for TagType {
    fn from(s: &str) -> Self {
        match s {
            "e" => TagType::Event,
            "p" => TagType::PubKey,
            e => TagType::Custom(e.to_owned()),
        }
    }
}

impl ToString for TagType {
    fn to_string(&self) -> String {
        match self {
            TagType::Event => "e".to_owned(),
            TagType::PubKey => "p".to_owned(),
            TagType::Custom(x) => x.to_owned(),
        }
    }
}

#[derive(Debug, Clone)]
pub struct Tag {
    pub typ: TagType,
    pub value: Addr,
    pub relayer_url: Option<String>,
    pub extra: Vec<String>,
}

impl Tag {
    pub fn new<T: Into<TagType>>(
        typ: T,
        value: Addr,
        relayer_url: Option<String>,
        extra: Vec<String>,
    ) -> Self {
        Self {
            typ: typ.into(),
            value,
            relayer_url,
            extra,
        }
    }
}

impl ser::Serialize for Tag {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let len = if self.relayer_url.is_some() {
            3 + self.extra.len()
        } else {
            2
        };
        let mut seq = serializer.serialize_seq(Some(len))?;
        seq.serialize_element(&self.typ.to_string())?;
        seq.serialize_element(&self.value.to_public_key())?;
        if let Some(addr) = self.relayer_url.as_ref() {
            seq.serialize_element(addr)?;
            for extra in self.extra.iter() {
                seq.serialize_element(extra)?;
            }
        }
        seq.end()
    }
}

impl<'de> Deserialize<'de> for Tag {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        let v: Vec<String> = Deserialize::deserialize(deserializer)?;
        if v.len() > 2 {
            return Err(de::Error::custom("expected array of two strings"));
        }
        let value: Addr = v[1]
            .as_str()
            .try_into()
            .map_err(|e: super::addr::Error| de::Error::custom(e.to_string()))?;
        let relayer_url = v.get(2).map(|x| x.to_owned());

        let extra = if v.len() > 3 {
            v[3..].to_owned()
        } else {
            vec![]
        };

        Ok(Self::new(v[0].as_str(), value, relayer_url, extra))
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn serialize_deserialize() {
        let json = "[\"p\",\"a0b0c0d0\"]";
        let tag: Tag = serde_json::from_str(&json).expect("valid json");
        assert_eq!(tag.typ, TagType::PubKey);
        assert_eq!(tag.value.to_public_key(), "a0b0c0d0".to_owned());
        assert!(tag.relayer_url.is_none());
        assert!(tag.extra.is_empty());
    }
}
